@@ -16,8 +16,8 @@
  * 
  * You should have received a copy of the GNU Library General Public
  * License along with the Gnome Library; see the file COPYING.LIB.  If not,
- * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  * 
  * Author: Soren Sandmann <sandmann@redhat.com>
  */
@@ -148,6 +148,7 @@ gboolean        gnome_rr_screen_set_dpms_mode      (GnomeRRScreen         *scree
 /* GnomeRROutput */
 guint32         gnome_rr_output_get_id             (GnomeRROutput         *output);
 const char *    gnome_rr_output_get_name           (GnomeRROutput         *output);
+const char *    gnome_rr_output_get_display_name   (GnomeRROutput         *output);
 gboolean        gnome_rr_output_is_connected       (GnomeRROutput         *output);
 int             gnome_rr_output_get_size_inches    (GnomeRROutput         *output);
 int             gnome_rr_output_get_width_mm       (GnomeRROutput         *outout);
